// ------------------------------------------------------------------
// depassement-tlb - Mesure de temps d'acces aleatoires sur des pages
//               memoire distinctes jusqu'a depassement du cache TLB.
//
// (c) 2011 Christophe BLAESS - Tous droits reserves.
//
// Exemple pour un article du blog http://www.blaess.fr/christophe/
// ------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/time.h>

typedef struct {
	int valeur;
	char padding[62]; // 64 - sizeof(int)
} entree_t;

int main(int argc, char * argv[])
{
	register int i, n;
	
	int nb_entrees;
	int nb_parcours;

	volatile entree_t ** entrees = NULL;

	struct timeval tv_debut;
	struct timeval tv_fin;
	long long int duree;

	if ((argc != 3)
	 || (sscanf(argv[1], "%d", & nb_entrees) != 1)
	 || (sscanf(argv[2], "%d", & nb_parcours) != 1)) {
		fprintf(stderr,"usage: %s nb_entrees nb_parcours\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	
	// On alloue une table de N pointeurs sur les entrees
	entrees = calloc(nb_entrees, sizeof(entree_t *));
	if (entrees == NULL) {
		perror("calloc");
		exit(1);
	}
	// Puis on alloue chaque entree sur une page specifique
	for (i = 0; i < nb_entrees; i ++) {
		entrees[i] = mmap(NULL, 4096, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
		if (entrees[i] == MAP_FAILED) {
			perror("mmap");
			exit(1);
		}
	}
	// La table est initialisee avec des -1
	for (i = 0; i < nb_entrees; i ++) {
		entrees[i]->valeur = -1;
	}
	n = 0;
	i = 0;
	// Remplir N-1 entrees (la derniere sert d'arret)
	while (n < nb_entrees-1) {
		// Remplir de maniere aleatoire avec le numero
		// d'une autre entree actuellement a -1.
		entrees[i]->valeur = rand() % nb_entrees;
		// Si elle n'est pas vide, rechercher la premiere
		// entree a -1
		while (entrees[entrees[i]->valeur]->valeur != -1) {
			entrees[i]->valeur ++;
			if (entrees[i]->valeur == nb_entrees)
				entrees[i]->valeur = 0;
		}
		i = entrees[i]->valeur;
		n++;
	}
	// La derniere entree contient toujours -1

	gettimeofday(& tv_debut, NULL);
	for (n = 0; n < nb_parcours; n ++) {
		i = 0;
		while(i != -1) {
			i=entrees[i]->valeur;
		}
	}
	gettimeofday(& tv_fin, NULL);
	duree  = tv_fin.tv_sec - tv_debut.tv_sec;
	duree *= 1000000;
	duree += tv_fin.tv_usec - tv_debut.tv_usec;


	// Duree en centiemes de nanosecondes
	duree  = duree  * 100000 / (nb_entrees * nb_parcours);
	
	fprintf(stdout, "%3lld.%02lld\n", duree/100, duree%100);
	return 0;
}

